(* Main.sml - Das Herz des Proof Assistenten
 *
 * In einer Hauptschleife werden die Standart Eingabe abgefragt, 
 * Ausdruecke geparst und durch Aufrufen der entsprechenden 
 * Subroutinen goals modifiziert.
 *)

local open BasicIO Nonstdio Data in
exception fileEnd 
exception exitProgram

fun welcome() = print ("NED Version 1.3\nType \"exit\" to quit " ^
		       "the program\n\"help\" gives more information\n\n"); 

fun errorpos i = "" (* (implode (List.tabulate (i+3, (fn n => #" ")))) ^ "^" *)

fun createLexerStream (is : instream) =
  Lexing.createLexer (fn buff => fn n => Nonstdio.buff_input is buff 0 n)

(* counts the Lines and chars*)
val currline = ref 1
val currchar = ref 0
val prompt   = ref true

(* Initiales,Backup und Save Goal *)
val initgoal = Goal.init_goal()
val backup   = ref (Goal.init_goal())  

(* returns (goal,savegoal) *)
(* Take current goal in history, if the last proof has finished, set the backup goal *)
fun getGoal(goalxs) =
    let val (headgoal, _) = hd(goalxs) 
	val nr = Goal.get_nr_of_goals(headgoal)
    in  
	if ( (nr > 0) andalso (Goal.get_first_goal(headgoal) = (C "PROOF DONE")))
	    then (Goal.init_goal_env(!backup), (!backup) ) 
	else (headgoal, initgoal)
    end


(* ------------------------------------
 *            Action Handler 
 * ------------------------------------ *)

fun ActionHandler (result as (CmdS(c))) goalxs = 
 let 
     val (goal,_) = getGoal(goalxs)
 (* ND-KOMMANDOS *)
 in   ( case c of 
       POSCMD i       => (Position.main(goal,i); goalxs )
     | HELPCMD        => (Hilfe.main() ; goalxs )
     | PRINTCMD       => (Goal.print_goal(goal); goalxs )
     | SHOWLABELSCMD  => (Visual.set_showlabels(); goalxs)
     | LATEXCMD s     => (LatexOutput.main((goalxs), s); goalxs) 
     | INTROCMD(m)    => (( (Intro.main(goal,m)), result)       :: goalxs )
     | ELIMCMD(m)     => (( (Elim.main (goal,m)), result)       :: goalxs )
     | REMOVECMD(d,i) => (( (Remove.main(goal,(d,i))), result)  :: goalxs )
     | MOVECMD(d)     => (( (Move.main (goal, d )), result)     :: goalxs )
     | SHIFTCMD(xd,yd,i) =>(( (Shift.main(goal, xd, (yd, i) )), result) ::goalxs ) 
     | GOALSHIFTCMD(i)=> (( Shift.goalmain(goal,i), result )    :: goalxs )
     | SIMPLIFYCMD    => (( (Bool.main(goal) ), result)         :: goalxs )
     | LAMCMD         => (( (Lambda.main(goal) ), result)       :: goalxs )
     | APPLYCMD(args) => (( (Apply.main(goal, args )), result)  :: goalxs )
     (* Wende die Cut-Regel an *)
     | CUTCMD (f)     =>  ( ( (Cut.main(goal, f)), result) :: goalxs )
     (* Instanziierung der Gammavariablen i mit Ausdruck f - instance i = f *)
     | INSTANCECMD(i,f) =>( (( Instance.main(goal,i,f) ), result) :: goalxs )
     | RESETCMD       => (print("> RESET DONE\n"); 
			  currline := 1; currchar := 0;
			  backup := (initgoal);
			  ((initgoal , EmptyS)::nil) ) 
     | UNDOCMD(i)     => if (List.length(goalxs) > i ) 
			     then let
				      val newgoalxs = List.drop(goalxs,i)
				      val (newfirst,_) = hd(newgoalxs)
				  in
				      (print "> UNDO successful...\n";
				       Goal.print_goal(newfirst); 
				       newgoalxs )
				  end
			 else raise Error ("Nothing to UNdo\n")
     | EXITCMD     => (print ("Bye\n") ; raise exitProgram)
			     ) (* end of case c *)
 end (* of let *)

(* Gleichheitskommandos *)
 | ActionHandler (result as (CmdEqS(c))) goalxs  = 
 let 
     val (goal,_) = getGoal(goalxs)
 in	  ( case c of
		REFEQ          => ( (RefSymTrans.reflexiv(goal), result)    ::goalxs) 
	      | SYMEQ(d)       => ( (RefSymTrans.symmetrie(goal,d), result) ::goalxs)
	      | TRANSEQ (f)    => ( (RefSymTrans.trans(goal,f), result)     ::goalxs)
	      | REWEQ(args)    => (((Rewriting.main(goal, args)), result)   ::goalxs)
	      | INSERTEQ(args) => ( ((Insert.main(goal,args)), result)      ::goalxs)
		    ) (* end of case *)
 end

(* Erweitere Variablenumgebung - var <name>:typ *)
  | ActionHandler (result as (LetVarS(s,t))) goalxs = 
 let                    
     val (goal,_) = getGoal(goalxs)
     val nr_active_goals = Goal.get_nr_of_goals(goal)
     val newgoal = 
	 if (nr_active_goals > 0 ) 
	     then raise Error ("You can't add new variables within a proof\n")
	 else Goal.update_ftypeenv (goal, s, t)
 in
     (print ("> var "^ s ^ ": " ^ Visual.ty2string(t) ^"\n" );
      ((newgoal, result)::goalxs) )
 end 

  (* Erweitere Konstantenumgebung - const <name>:typ *)
  | ActionHandler (result as (LetConstS(s,t))) goalxs =
 let     
     val (goal,_) = getGoal(goalxs)
     val nr_active_goals = Goal.get_nr_of_goals(goal)
     val newgoal = 
	 if (nr_active_goals > 0 ) 
	     then raise Error ("You can't add new constants within a proof\n")
	 else Goal.update_ctypeenv (goal, s, t)
 in
     (print ("> const "^ s ^ ":" ^ Visual.ty2string(t) ^ "\n");
      ((newgoal, result)::goalxs) )
 end

  (* Deklaration von Definitionen - def <name>:<ausdruck>    hier: mode=CLASSIC
   *                              lemma <name>:<ausdruck>          mode=LEMMA
   *                              axiom <name>:<ausdruck>          mode=AXIOM
   *)	       
  | ActionHandler (result as (DefS(s,f,mode))) goalxs = 
 let
     val (goal,_) = getGoal(goalxs)
     val nr_active_goals = Goal.get_nr_of_goals(goal)
     val name = case mode of CLASSIC => "def" | LEMMA => "lemma" | AXIOM => "axiom"
     val newgoal =
	 if (nr_active_goals > 0 ) 
	     then 
		 raise Error ("You can't add new definitions within a proof\n")
	 else 
	     Goal.new_definition(goal, s, f, mode ) 
 in
     (print ("> " ^ name ^ " " ^ s ^ " saved \n");
      ( (newgoal, result)::goalxs) )
 end

  (* Typneudeklaration - type <name> *)
  | ActionHandler (result as (TypS(s))) goalxs   =
 let
     val (goal,_) = getGoal(goalxs)
     val nr_active_goals = Goal.get_nr_of_goals(goal)
     val newgoal = 
	 if ( nr_active_goals > 0 ) 
	     then raise Error ("You can't add new types within a proof\n")
	 else Goal.new_type(goal, s)
 in
     (print ("> type " ^ s ^ "\n") ; 
      ((newgoal, result)::goalxs) )
 end

  (* Neues goal - goal <ausdruck> *)
  | ActionHandler (result as (GoalS(f))) goalxs  = 
 let 
     val (goal,_) = getGoal(goalxs)
     val nr_goals = Goal.get_nr_of_goals(goal)
 in 
     if ( nr_goals = 0 ) then (* Das erste Goal wurde deklariert *)
	 let val newgoal = Goal.first_goal(goal, f)
	 in
	     if Goal.typecheck(newgoal) then
		 (
		  (* speichere das Goal, das alle Var./Const/Ax/Lemma
		   * Deklarationen enthaelt, ohne Goalspezifikation 
		   *)
		  backup := newgoal;						  
		  Goal.print_goal(newgoal) ; 
		  ((newgoal, result)::goalxs))
	     else raise Error ("Ill-Typed Formula ( must have base type bool )!\n") 
	 end
     else ( Goal.print_goal(goal) ; 
	   raise Error ("There are already active goals \n"))
 end
	       
  (* Das bewiesene Goal soll unter <s> gespeichert werden *)
  | ActionHandler (result as (SaveS(s))) goalxs  = 
 let
     val (goal, savegoal) = getGoal(goalxs)
     val check = Goal.get_nr_of_goals(savegoal)
     val newgoal = 
	 if (check > 0) 
	     then let val axpter = 
		 ( case (Goal.get_first_goal(savegoal)) of
		       IMP(C "TRUE", p) => p
		     | _ => raise Error ("Save Main Internal Error \n") )
		  in
		      (print ("Lemma <" ^ s ^ "> saved \n") ;
		       Goal.save_goal(goal, s, axpter)      )
		  end
	 else raise Error ("There is no goal to save \n")
 in 
     ((newgoal, result)::goalxs)
 end

  (* fuer direkte Formeleingabe *)
  | ActionHandler (ForS(f)) _  = raise Error ("Please use <goal formula> for input \n")
		   
  (* fuer leere inputs *)
  | ActionHandler (EmptyS) goalxs =  (prompt:=false; goalxs)
		   
  (* Parser/Lexer Errors *)
  | ActionHandler (result as (ErrorS(msg))) goalxs = 
    let
	val goal =( case goalxs of 
		 ((g,_)::xs) => g 
	       | _ => raise Error ("Internal Error Main.sml \n"))
    in  ( print msg; ((goal, result)::goalxs) )
    end
		   
  | ActionHandler (FileEndS) _ = raise fileEnd
		   
(* end of ActionHandler *)
		   
				     

(* Entry Point of NED *)
fun start(args) = 
    let 
	val _ = welcome()
	datatype input_mode = FILESTREAM | STDINSTREAM 
	val (stream, mode) = 
	    ( case args of 
		  (s::nil) => ( (( Nonstdio.open_in_bin s ), FILESTREAM ) 
			       handle SysErr _ => 
				   (print ("-- Error: File <"^s^"> not found --\n\n"); 
				    (std_in, STDINSTREAM )) )
		    | _ => (std_in,STDINSTREAM)                                         ) 
        
        val buffer  = createLexerStream (stream)

	val currbuf  = ref (buffer)
	val currmode = ref (mode)

	val initentry = (initgoal, EmptyS)

	fun switch_buffer() = if ((!currmode) = FILESTREAM) then 
	                        ( BasicIO.close_in stream;
				  currbuf := createLexerStream (std_in);
				  currline := 1 ;
				  currmode := STDINSTREAM )
			       else () 
					
	 
	(* Haupt Schleife mit Argument goalxs Liste (History) *)
	fun mainloop goalxs : unit = 
	    ( let 
		  val lexbuf = (!currbuf)
		  val _ = if (!currmode) = FILESTREAM orelse 
			      not (!prompt) 
			      then (prompt:=true) else print ("- ")
		      
		  (* result enthaelt parseexpr *)
		  val result = ( Parser.Main Lexer.Token lexbuf )
				 
		      (* Fehlerbehandlung Parser/Lexer *)
		      handle Lexer.LexicalError (s,l1,l2) 
		      => let val name = Lexing.getLexeme lexbuf
		             val l = (!currline)
			     val c = (!currchar)
			 in (switch_buffer(); 
			     currline := (!currline) - 1;
			     (ErrorS (  "\n> Lexer Error : " ^ s 
				      ^ "\n  BAD SYMBOL : " ^ name 
				      ^ "\n  in Line : " ^ Int.toString(l)
				      ^ "  at Pos : " ^ Int.toString(l1-c) 
				      ^ "-" ^ Int.toString(l2-c) ^ "\n" )) )
			 end
 		      
		    | Parsing.ParseError ob => 
		      let val pos1 = Lexing.getLexemeStart lexbuf
			  val pos2 = Lexing.getLexemeEnd lexbuf
			  val name = Lexing.getLexeme lexbuf
			  val l = (!currline)
			  val c = (!currchar)
			      
		      in  (switch_buffer();
			   currline := (!currline) - 1;
			   (ErrorS  (  "\n> Parser Error"
				    ^ "\n  BAD SYMBOL : " ^ name 
				    ^ "\n  in Line : "^ Int.toString(l)
				    ^ "  at Pos : " ^ Int.toString(pos1-c)
				    ^ "-" ^ Int.toString(pos2-c) ^ "\n")) )
			
		      end 
		    | Error msg => 
		      let
			  val l = (!currline)
		      in (switch_buffer(); 
			  currline := (!currline) - 1;
			  ErrorS ("> " ^ msg ^ "\n  in Line : " ^ Int.toString(l) ^ "\n"))    (* Fehler *) 
		      end
		
	
		  val (goal,_) = getGoal(goalxs)
		  val poschar  = Lexing.getLexemeEnd lexbuf
		  val newgoalxs = ( ActionHandler result goalxs )
		      handle fileEnd => (print ("> Input-File successfully loaded \n");
					 switch_buffer(); goalxs ) 
 			   | Error msg => (print ("> Error : " ^ msg 
						  ^ "  in Line : " ^ Int.toString(!currline) ^ "\n");
					   switch_buffer();
					   ((goal, ErrorS "FEHLER")::goalxs) )     (* Fehler *)  
	    
	      in
		 ( currline := (!currline) + 1 ;
		   currchar := poschar ;
		   mainloop newgoalxs )
	      end
	  )  handle exitProgram => () 

    
    in
	mainloop (initentry::nil)
    end


(*-NED-*) val _ = start(CommandLine.arguments())


end; (* end of local open *)
